/*
 * Copyright (c) 2010-2015 TradertoolsFX
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#property copyright "Copyright (c) 2010-2015 TradertoolsFX"
#property link      "http://www.tradertools-fx.com"

// Indices
#define IFIB 0 // Fibo object index
#define IRET 1 // Retraced zone object index
#define IUNRET 2 // Unretraced zone object index

/* Structure to hold some info about chart objects
 * in RetracementDrawer as well as
 * some small functions for conveniency. */
struct ObjInfo {
    string name; // The name of the object used to identify it uniquely.
    ENUM_OBJECT type;
    color color_; // Only used by the rectangle objects.

    /* Copy structure members from other. */
    void copy(ObjInfo &other)
    {
        name = other.name;
        type = other.type;
        color_ = other.color_;
    }
    void operator=(ObjInfo &other) { copy(other); }

    /* Return the object's first price coordinate. */
    double p1(void) { return ObjectGet(name, OBJPROP_PRICE1); }
    /* Return the object's second price coordinate. */
    double p2(void) { return ObjectGet(name, OBJPROP_PRICE2); }
    /* Return true if one of the object's  price coordinates is not equal
     * to the respective price p1 or p2. */
    bool update_required(double p1, double p2) { return (p1 != p1()) || (p2 != p2()); }
};

/* Structure to hold important bar indices. */
struct BarIndex {
    int ihighest, ilowest, istart, iend;
};

/* Class to manage drawing and updating of three chart objects
 * that convey information about a retracement -
 * Fibonacci Retracement, Retraced Zone, Unretraced Zone. */
class RetracementDrawer {
public:
    RetracementDrawer(double &levels[],
                      int num_levels,
                      const string &prefix,
                      color retraced_color,
                      color unretraced_color);
    ~RetracementDrawer(void);

    void set_range(int idx_start, int idx_end);

    int draw_fibo(bool redraw);
    int update_fibo(void);

    int draw_retraced_zone(bool redraw);
    int update_retraced_zone(void);

    int draw_unretraced_zone(bool redraw);
    int update_unretraced_zone(void);
private:
    double m_levels[];
    int m_num_levels;
    ObjInfo m_obj[3];
    BarIndex m_bar;

    void set_ilowest(void);
    void set_ihighest(void);

    void set_obj_names(string prefix);
    void set_obj_types(void);

    void set_fibo_objprops(const string &name);
    void set_fibo_levels(const string &name);

    datetime time(int point, int iobj);
    datetime fibo_time(int point);
    datetime retraced_time(int point);
    datetime unretraced_time(int point);

    double price(int point, int iobj);
    double fibo_price(int point);
    double retraced_price(int point);
    double unretraced_price(int point);

    int imax_retraced(void);
    double max_retraced_price(void);

    bool is_trend_up(void);

    bool fibo_update_required(void);
    bool retraced_update_required(void);
    bool unretraced_update_required(void);

    int update(int iobj);
    int draw(int iobj, bool redraw);
};

RetracementDrawer::RetracementDrawer(double &levels[],
                                     int num_levels,
                                     const string &prefix,
                                     color retraced_color,
                                     color unretraced_color)
{
    ArrayCopy(m_levels, levels);
    m_num_levels = num_levels;
    m_obj[IRET].color_ = retraced_color;
    m_obj[IUNRET].color_ = unretraced_color;
    set_obj_names(prefix);
    set_obj_types();
}

RetracementDrawer::~RetracementDrawer()
{
    for (int i = 0; i < ArraySize(m_obj); i++) {
        ObjectDelete(m_obj[i].name);
    }
}

/* Set the bar range to be considered in the retracement given the
 * start and ending bar index, and set the indices for the low and high
 * over that range. */
void RetracementDrawer::set_range(int istart, int iend)
{
    m_bar.istart = istart;
    m_bar.iend = iend;

    set_ilowest();
    set_ihighest();
}

/* Set ilowest to the index of the bar with the lowest low. */
void RetracementDrawer::set_ilowest(void)
{
    m_bar.ilowest = iLowest(NULL,
                            0,
                            MODE_LOW,
                            m_bar.iend + m_bar.istart,
                            m_bar.istart);
}

/* Set ihighest to the index of the bar with the highest high. */
void RetracementDrawer::set_ihighest(void)
{
    m_bar.ihighest = iHighest(NULL,
                              0,
                              MODE_HIGH,
                              m_bar.iend + m_bar.istart,
                              m_bar.istart);
}

/* Set all the object names in m_obj. */
void RetracementDrawer::set_obj_names(string prefix)
{
    int sz = ArraySize(m_obj);
    for (int i = 0; i < sz; i++) {
        m_obj[i].name = StringConcatenate(prefix, i);
    }
}

/* Set all the object types in m_obj. */
void RetracementDrawer::set_obj_types(void)
{
    m_obj[IFIB].type = OBJ_FIBO;
    m_obj[IRET].type = OBJ_RECTANGLE;
    m_obj[IUNRET].type = OBJ_RECTANGLE;
}

/* Set fibo chart object properties given the object's name. */
void RetracementDrawer::set_fibo_objprops(const string &name)
{
    ObjectSet(name, OBJPROP_LEVELWIDTH, 1);
    ObjectSet(name, OBJPROP_LEVELSTYLE, 0);
    ObjectSet(name, OBJPROP_COLOR, clrYellow);
}

/* Set fibo chart object's retracement levels given the object's name. */
void RetracementDrawer::set_fibo_levels(const string &name)
{
    string level, desc;

    ObjectSet(name, OBJPROP_FIBOLEVELS, m_num_levels);
    for (int i = 0; i < m_num_levels; i++) {
        ObjectSet(name, OBJPROP_FIRSTLEVEL + i, m_levels[i]);
        // Set the fibo level description
        level = DoubleToStr(m_levels[i], 3);
        desc = StringConcatenate(level, " - %$");
        ObjectSetFiboDescription(name, i, desc);
    }
}

/* Return the index of the bar with price that is the max retraced. */
int RetracementDrawer::imax_retraced(void)
{
    int res;

    if (is_trend_up()) {
        res = iLowest(NULL,
                      0,
                      MODE_LOW,
                      m_bar.ihighest + m_bar.istart,
                      m_bar.istart);
    } else {
        res = iHighest(NULL,
                       0,
                       MODE_HIGH,
                       m_bar.ilowest + m_bar.istart,
                       m_bar.istart);
    }

    return res;
}

double RetracementDrawer::max_retraced_price(void)
{
    int imax = imax_retraced();
    if (is_trend_up()) {
        return Low[imax];
    } else {
        return High[imax];
    }
}

bool RetracementDrawer::is_trend_up(void)
{
    return (m_bar.ihighest < m_bar.ilowest);
}

/* Draw chart object given the
 * index iobj (IFIB, IRET, or IUNRET).
 * Return 0 on success, -1 if error, -2 if already exists. */
int RetracementDrawer::draw(int iobj, bool redraw)
{
    ObjInfo obj = m_obj[iobj];
    bool success;

    if (redraw ||
            ObjectFind(obj.name) == -1) {
        ObjectDelete(obj.name);
        success = ObjectCreate(obj.name,
                               obj.type,
                               0,
                               time(0, iobj),
                               price(0, iobj),
                               time(1, iobj),
                               price(1, iobj));
        if (!success) {
            return -1;
        }

        if (obj.type == OBJ_FIBO) {
            set_fibo_objprops(obj.name);
            set_fibo_levels(obj.name);
        } else {
            ObjectSet(obj.name, OBJPROP_COLOR, obj.color_);
        }
    } else { // Object already exists
        return -2;
    }

    return 0;
}

int RetracementDrawer::draw_fibo(bool redraw)
{
    return draw(IFIB, redraw);
}

int RetracementDrawer::draw_retraced_zone(bool redraw)
{
    return draw(IRET, redraw);
}

int RetracementDrawer::draw_unretraced_zone(bool redraw)
{
    return draw(IUNRET, redraw);
}

/* Update chart object at index iobj if necessary.
 * Return 0 on success. */
int RetracementDrawer::update(int iobj)
{
    ObjInfo obj = m_obj[iobj];
    int ipoint = 0;
    bool success;

    // Check if update required and assign value to success flag.
    if (success = obj.update_required(price(0, iobj), price(1, iobj))) {
        // Move both points
        while (success && ipoint < 2) {
            success = ObjectMove(obj.name,
                                 ipoint,
                                 time(ipoint, iobj),
                                 price(ipoint, iobj));
            ipoint++;
        }

        if (!success) {
            return -1;
        }
    }

    return 0;
}

int RetracementDrawer::update_fibo(void)
{
    return update(IFIB);
}

int RetracementDrawer::update_retraced_zone(void)
{
    return update(IRET);
}

int RetracementDrawer::update_unretraced_zone(void)
{
    return update(IUNRET);
}

bool RetracementDrawer::fibo_update_required(void)
{
    return m_obj[IFIB].update_required(fibo_price(0),
                                       fibo_price(1));
}

bool RetracementDrawer::retraced_update_required(void)
{
    return m_obj[IRET].update_required(retraced_price(0),
                                       retraced_price(1));
}

bool RetracementDrawer::unretraced_update_required(void)
{
    return m_obj[IUNRET].update_required(unretraced_price(0),
                                         unretraced_price(1));
}

/* Return the time coordinate that an object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point and object index iobj.
 * Note! The point index starts at 0, so 0 corresponds
 * to the time for coordinate 1. */
datetime RetracementDrawer::time(int point, int iobj)
{
    switch (iobj) {
    case IFIB:
        return fibo_time(point);
    case IRET:
        return retraced_time(point);
    case IUNRET:
        return unretraced_time(point);
    default:
        return -1;
    }
}

/* Return the time coordinate that the fibo object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
datetime RetracementDrawer::fibo_time(int point)
{
    datetime time[2];
    datetime time_low = Time[m_bar.ilowest];
    datetime time_high = Time[m_bar.ihighest];

    if (is_trend_up()) {
        time[0] = time_low;
        time[1] = time_high;
    } else {
        time[0] = time_high;
        time[1] = time_low;
    }

    return time[point];
}

/* Return the time coordinate that the retraced object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
datetime RetracementDrawer::retraced_time(int point)
{
    datetime time[2];

    time[0] = fibo_time(1);
    time[1] = Time[imax_retraced()];

    return time[point];
}

/* Return the time coordinate that the unretraced object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
datetime RetracementDrawer::unretraced_time(int point)
{
    datetime time[2];

    time[0] = fibo_time(1);
    time[1] = Time[imax_retraced()];

    return time[point];
}

/* Return the price coordinate that an object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point and object index iobj.
 * Note! The point index starts at 0, so 0 corresponds
 * to the price for coordinate 1. */
double RetracementDrawer::price(int point, int iobj)
{
    switch (iobj) {
    case IFIB:
        return fibo_price(point);
    case IRET:
        return retraced_price(point);
    case IUNRET:
        return unretraced_price(point);
    default:
        return -1;
    }
}

/* Return the price coordinate that the fibo object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
double RetracementDrawer::fibo_price(int point)
{
    double price[2];
    double low = Low[m_bar.ilowest];
    double high = High[m_bar.ihighest];

    if (is_trend_up()) {
        price[0] = low;
        price[1] = high;
    } else {
        price[0] = high;
        price[1] = low;
    }

    return price[point];
}

/* Return the price coordinate that the retraced object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
double RetracementDrawer::retraced_price(int point)
{
    double price[2];

    price[0] = fibo_price(1);
    price[1] = max_retraced_price();

    return price[point];
}

/* Return the price coordinate that the unretraced object should be
 * set to (NOT necessarily the actual object's coordinate *yet*)
 * given the point index point. */
double RetracementDrawer::unretraced_price(int point)
{
    double price[2];

    price[0] = max_retraced_price();
    price[1] = fibo_price(0);

    return price[point];
}
